package Stanford::Vemail;

use DBI;
use Net::LDAPapi;
use Readonly;
use Time::HiRes;
use YAML::Tiny;

use Stanford::Orange::Util qw/kerberos_ticket_cache_no_afs/;

Readonly my $CONFIG_FILE => '/etc/vemail/config.yaml';

###########################################################################
###########################################################################

# Non object methods

sub valid_host {
    ## no critic (ValuesAndExpressions::ProhibitMagicNumbers);

    my ($hn) = @_;
    my @ip = split(/[.]/xsm, $hn);

    # check for a ip literal
    if (scalar(@ip) == 4) {
        my $okay = 1;
        for my $i (0..3) {
            if ($ip[$i] !~ /^\d+$/xsm) { $okay = 0; }
            if ($ip[$i] < 0)           { $okay = 0; }
            if ($ip[$i] > 255)         { $okay = 0; }
            if (!$okay) {
                last;
            }
        }
        if ($okay) { return 1; }
    }

    # If we don't get 4 octets it is because it is a bad hostname
    my $raw_addr = (gethostbyname($hn))[4];
    my @octets = unpack('C4', $raw_addr);
    if (!scalar(@octets)) {
        return 0;
    }

    # Don't use the hostname, but might need it sometime
    my $ip = join(q{.}, @octets);
    return 1;
}

sub exit_with_error {
    my ($msg) = @_ ;

    print "ERROR: $msg\n";
    exit 1;
}

sub read_pwfile {
    my ($pw_file) = @_ ;

    open(my $FH, '<', $pw_file) ;
    my $line = <$FH> ;
    chomp($line);
    close($FH) ;

    return $line ;
}

###########################################################################
###########################################################################

sub new {
    my ($proto, %args) = @_ ;
    my $self = {} ;
    my $class = ref($proto) || $proto ;

    $self->{'config_file'}  = $args{'config_file'} || $CONFIG_FILE ;
    $self->{'debug'}        = $args{'debug'}       || 0 ;
    $self->{'debug_time'}   = $args{'debug_time'}  || undef ;
    $self->{'dryrun'}       = $args{'dryrun'}      || 0 ;

    # CONF is a hash reference.
    $self->{'CONF'}         = $args{'CONF'}        || {} ;

    # The LDAP connection objects.
    $self->{'ldap_read'}  = undef ;
    $self->{'ldap_write'} = undef ;

    # A DBI connect object.
    $self->{'DB'} = undef ;

    bless($self, $class) ;

    $self->initialize() ;

    return $self ;
}

sub initialize {
    my $self = shift ;

    $self->set_debug_time() ;

    return ;
}

sub config_file {
    my $self = shift ;
    my ($config_file) = @_ ;

    if (defined($config_file)) {
        $self->{'config_file'} = $config_file ;
    }

    return $self->{'config_file'} ;
}

sub debug {
    my $self = shift ;
    my ($debug) = @_ ;

    if (defined($debug)) {
        $self->{'debug'} = $debug ;
    }

    return $self->{'debug'} ;
}

sub dryrun {
    my $self = shift ;
    my ($dryrun) = @_ ;

    if (defined($dryrun)) {
        $self->{'dryrun'} = $dryrun ;
    }

    return $self->{'dryrun'} ;
}

sub CONF {
    my $self = shift ;
    my ($CONF) = @_ ;

    if (defined($CONF)) {
        $self->{'CONF'} = $CONF ;
    }

    return $self->{'CONF'} ;
}

sub ldap_read {
    my $self = shift ;
    my ($ldap_read) = @_ ;

    if (defined($ldap_read)) {
        $self->{'ldap_read'} = $ldap_read ;
    }

    return $self->{'ldap_read'} ;
}

sub ldap_write {
    my $self = shift ;
    my ($ldap_write) = @_ ;

    if (defined($ldap_write)) {
        $self->{'ldap_write'} = $ldap_write ;
    }

    return $self->{'ldap_write'} ;
}

sub DB {
    my $self = shift ;
    my ($DB) = @_ ;

    if (defined($DB)) {
        $self->{'DB'} = $DB ;
    }

    return $self->{'DB'} ;
}

sub debug_time {
    my $self = shift ;
    return $self->{'debug_time'} ;
}

# Update debug time to right now.
sub set_debug_time {
    my $self = shift ;

    my $now = Time::HiRes::time();
    $self->{'debug_time'} = $now;

    return ;
}

sub elapsed_time {
    my $self = shift ;

    my $now = Time::HiRes::time();
    my $elapsed = $now - $self->debug_time();

    return $elapsed ;
}


sub dbg {
    my $self = shift ;
    my ($msg) = @_;

    if ($self->debug()) {

        my $now     = Time::HiRes::time();
        my $elapsed = $self->elapsed_time();

        my $now_rounded     = sprintf('%.4f', $now);
        my $elapsed_rounded = sprintf('%.3f', $elapsed);

        print "$now_rounded ($elapsed_rounded) $msg\n";
        $self->set_debug_time() ;
    }

    return;
}

sub read_config {
    my $self = shift ;

    my $CONF = $self->CONF() ;

    ## STEP 1: Set some defaults.
    $CONF->{'add'}            = 1 ;
    $CONF->{'db_conf'}        = '/etc/mysql.secret';
    $CONF->{'container_base'} = 'cn=applications,dc=stanford,dc=edu';
    $CONF->{'debug'}          = 0 ;
    $CONF->{'delete'}         = 1 ;
    $CONF->{'use_default_mail_domain'} = 0;
    $CONF->{'verbose'}        = 0;

    if ($debug) {
        $self->debug(1);
    }

    # Table names.
    $CONF->{'alias_table'}     = 'aliases';
    $CONF->{'container_table'} = 'containers';

    # alias table field mappings
    my $alias_table_mappings = {
        'mail'            => 'mail',
        'cn'              => 'container',
        'description'     => 'description',
        'owner',          => 'owner',
        'sumaildrop',     => 'sumaildrop',
        'suadministrator' => 'suadministrator',
    };
    $CONF->{'alias_attr_mapping'} = $alias_table_mappings;

    # Container table fields to ldap attribute mapping
    $CONF->{'container_table'} = 'containers';

    my $container_table_mappings = {
        'cn'          => 'container',
        'description' => 'workgroup',
    } ;
    $CONF->{'container_attr_mapping'} = $container_table_mappings;

    # The database connection settings.
    $CONF->{'prod_port'} = '3306' ;

    $self->dbg('finished setting configuration defaults');

    ## STEP 2: Read config.yaml and update the $CONF has reference.
    my $config_file = $self->config_file();
    $self->dbg("reading config file $config_file");
    my $yaml = YAML::Tiny->read($config_file);
    my $config = $yaml->[0];

    ### ###    ### ###    ### ###    ### ###    ### ###    ### ###
    my $read_and_set_fref = sub {
        my ($primary) = @_ ;
        $self->dbg("reading $primary settings") ;
        $CONF->{$primary} = $config->{$primary};
        for my $secondary (keys %{$config->{$primary}}) {
            my $tmp = $config->{$primary}->{$secondary};
            $self->dbg("set ${primary}:${secondary} ($tmp)");
        }
    } ;
    ### ###    ### ###    ### ###    ### ###    ### ###    ### ###

    ## kerberos
    if (exists($config->{'kerberos'})) {
        $read_and_set_fref->('kerberos') ;
    } else {
        $self->dbg('no kerberos settings to read') ;
    }

    ## ldap
    if (exists($config->{'ldap'})) {
        $read_and_set_fref->('ldap') ;
    } else {
        $self->dbg('no ldap settings to read') ;
    }

    ## db
    if (exists($config->{'db'})) {
        $read_and_set_fref->('db') ;
    } else {
        $self->dbg('no db settings to read') ;
    }

    ## template
    if (exists($config->{'template'})) {
        $self->dbg('reading template setting') ;
        my $template = $config->{'template'};
        $CONF->{'template'} = $template;
        $self->dbg("set template ($template)");
    } else {
        $self->dbg('no template settings to read') ;
    }

    ## admins
    if (exists($config->{'admins'})) {
        $self->dbg('reading admins setting') ;
        my $admins = $config->{'admins'};
        $CONF->{'admins'} = $admins;
        my $admins_str = join(',', @$admins) ;
        $self->dbg("set admins ($admins_str)");
    } else {
        $self->dbg('no admins settings to read') ;
    }

    ## alias table attribute mappings
    if (exists($config->{'alias_attr_mapping'})) {
        $self->dbg('alias_attr_mapping') ;
        $CONF->{'alias_attr_mapping'} = $config->{'alias_attr_mapping'};
        $self->dbg('set alias_attr_mapping');
    } else {
        $self->dbg('no alias_attr_mapping settings to read') ;
    }

    ## container table attribute mappings
    if (exists($config->{'container_attr_mapping'})) {
        $self->dbg('container_attr_mapping') ;
        $CONF->{'container_attr_mapping'} = $config->{'container_attr_mapping'};
        $self->dbg('set container_attr_mapping');
    } else {
        $self->dbg('no container_attr_mapping settings to read') ;
    }


    # Do some validations.
    # ##     # ##     # ##     # ##     # ##     # ##     # ##     # ##     # ##
    my $check_setting_two_level_fref = sub {
        my ($primary, $secondary) = @_ ;

        if (!$CONF->{$primary}->{$secondary}) {
            exit_with_error("$primary:$secondary setting missing") ;
        } else {
            $self->dbg("$primary:$secondary passed validation")
        }
    };
    my $check_setting_one_level_fref = sub {
        my ($primary) = @_ ;

        if (!$CONF->{$primary}) {
            exit_with_error("$primary setting missing") ;
        } else {
            $self->dbg("$primary passed validation")
        }
    };
    # ##     # ##     # ##     # ##     # ##     # ##     # ##     # ##     # ##
    $check_setting_two_level_fref->('kerberos', 'keytab');
    $check_setting_two_level_fref->('kerberos', 'tgt');
    $check_setting_two_level_fref->('ldap',     'write_host');
    $check_setting_two_level_fref->('db',       'database');
    $check_setting_two_level_fref->('db',       'port');
    $check_setting_two_level_fref->('db',       'username');
    $check_setting_two_level_fref->('db',       'password_file');
    $check_setting_two_level_fref->('db',       'host');

    $check_setting_one_level_fref->('template');
    $check_setting_one_level_fref->('admins');

    $self->dbg('finished reading configuration file');

    return ;
}

sub database_connect {
    my $self = shift ;

    my $CONF = $self->CONF() ;

    my $DB = $CONF->{'db'};

    my $host     = $DB->{'host'} ;
    my $database = $DB->{'database'} ;
    my $port     = $DB->{'port'} ;
    my $username = $DB->{'username'} ;
    my $password = read_pwfile($DB->{'password_file'}) ;

    my $dsn =   "DBI:mysql:host=$host;"
              . "database=$database;"
              . "host=$host;"
              . "port=$port";
    $self->dbg("about to connect to DB using dsn '$dsn'");

    $DBH = DBI->connect($dsn, $username, $password)
        or exit_with_error("unable to connect to database: $DBI::errstr");

    $self->DB($DBH) ;

    $self->dbg('successfully connected to database') ;

    return ;
}

sub database_disconnect {
    my $self = shift ;

    my $DBH = $self->DB() ;

    if ($DBH) {
        $DBH->disconnect();
    }

    return ;
}

# Set up a Kerberos context.
sub kinit {
    my $self = shift ;

    my $CONF = $self->CONF() ;

    my $keytab = $CONF->{'kerberos'}->{'keytab'} ;
    my $cache  = $CONF->{'kerberos'}->{'tgt'} ;

    return kerberos_ticket_cache_no_afs($keytab, $cache);
}


# ------------------------------------------------
# Connect to the LDAP.

sub ldap_write_connect {
    my $self = shift ;
    return $self->ldap_connect(1) ;
}

sub ldap_read_connect {
    my $self = shift ;
    return $self->ldap_connect(0) ;
}

# Pass the parameter $write as 1 if we want to connect to the
# master, 0 if we want to connect to the replicas.
sub ldap_connect {
    ## no critic (ValuesAndExpressions::ProhibitMagicNumbers);
    my $self = shift ;
    my ($write) = @_ ;

    my $CONF = $self->CONF() ;

    my $ldap_host ;
    if ($write) {
        $self->dbg('connecting to LDAP as WRITER');
        $ldap_host = $CONF->{'ldap'}->{'write_host'} ;
    } else {
        $self->dbg('connecting to LDAP as READER');
        $ldap_host = $CONF->{'ldap'}->{'read_host'} ;
    }

    if (!valid_host($ldap_host)) {
        exit_with_error("invalid host name $ldap_host");
    }

    # Bind to the server
    $self->dbg("binding to server $ldap_host");
    my $LDAP ;
    if (($LDAP = Net::LDAPapi->new($ldap_host)) == -1) {
        exit_with_error("Connection to $ldap_host failed");
    }
    $LDAP->sasl_parms(-mech=>'GSSAPI');

    my $status = $LDAP->bind_s(-type=>LDAP_AUTH_SASL);
    if ($status != LDAP_SUCCESS) {
        $LDAP->unbind;
        exit_with_error("AUTH error connecting to $ldap_host");
    }

    if ($write) {
        $self->ldap_write($LDAP);
        $self->dbg("successfully connected and bound to LDAP WRITE server $ldap_host");
    } else {
        $self->ldap_read($LDAP);
        $self->dbg("successfully connected and bound to LDAP READ server $ldap_host");
    }

    return;
}

sub ldap_unbind {
    my $self = shift ;

    my $ldap_read = $self->ldap_read() ;
    if ($ldap_read) {
        $ldap_read->unbind;
    }

    my $ldap_write = $self->ldap_write() ;
    if ($ldap_write) {
        $ldap_write->unbind;
    }

    return ;
}


1;
