#!/usr/bin/perl
#
# vemail -- Virtual Domain for Email management
#
# Written by Huaqing Zheng <morpheus@stanford.edu>
# Updated by Adam H. Lewenberg <adamhl@stanford.edu>

## no critic (CodeLayout::ProhibitParensWithBuiltins);
## no critic (Modules::RequireNoMatchVarsWithUseEnglish);

##############################################################################
# Site configuration
##############################################################################

use strict;
use warnings;
use autodie;

use Data::Dumper ;

use Carp;
use CGI;
use DBI;
use Email::Address;
use English;
use HTML::Template;
use Net::LDAPapi;
use Readonly;
use YAML::Tiny;

use Stanford::Vemail;
use Stanford::Orange::Util qw(
    kerberos_ticket_cache_no_afs
    run_command_improved
    slurp
    trim
);

my $VEMAIL = Stanford::Vemail->new();
my $LDAP;
my $CONF;

Readonly my $LDAP_CONNECT_FAILED    => -1;
Readonly my $NUMBER_OF_ROWS_TO_SHOW => 10;

##############################################################################
# Test Code
##############################################################################

# $ENV{'WEBAUTH_LDAP_SUPRIVILEGEGROUP'} = 'itservices:idg';

##############################################################################
# Helper routines
##############################################################################

## Read the configuration file
#sub read_config {
#
#    my $yaml = YAML::Tiny->read($CONFFILE);
#    my $config = $yaml->[0];
#
#    $CONF->{'template'}         = $config->{'template'};
#    $CONF->{'tgt'}              = $config->{'kerberos'}->{'tgt'};
#    $CONF->{'keytab'}           = $config->{'kerberos'}->{'keytab'};
#    $CONF->{'admins'}           = $config->{'admins'};
#    $CONF->{'db_host'}          = $config->{'db'}->{'host'};
#    $CONF->{'db_port'}          = $config->{'db'}->{'port'};
#    $CONF->{'db_database'}      = $config->{'db'}->{'database'};
#    $CONF->{'db_password_file'} = $config->{'db'}->{'password_file'};
#    $CONF->{'ldap_read_host'}   = $config->{'ldap'}->{'read_host'};
#    $CONF->{'ldap_write_host'}  = $config->{'ldap'}->{'write_host'};
#    $CONF->{'ldap_wg_base'}     = $config->{'ldap'}->{'wg_base'};
#
#    # Read the database password from its file.
#    $DB->{'username'} = $config->{'db'}->{'username'};
#    $DB->{'password'} = read_password($CONF->{'db_password_file'});
#
#    return;
#}
#
#sub read_password {
#    my ($password_file) = @_ ;
#
#    if (!$password_file) {
#        croak 'missing password_file parameter';
#    }
#
#    if (!(-f $password_file)) {
#        croak "no password_file '$password_file'";
#    }
#
#    return trim(slurp($password_file));
#
#}
#
#sub refresh_krb5_cache {
#    my ($keytab_file, $cache_file) = @_ ;
#
#    if (!$keytab_file) {
#        croak 'no keytab_file parameter passed' ;
#    }
#
#    if (!$cache_file) {
#        croak 'no cache_file parameter passed' ;
#    }
#
#    return kerberos_ticket_cache_no_afs($keytab_file, $cache_file);
#}
#
## Connect to the database
#sub db_connect {
#    my $dbh = DBI->connect('DBI:mysql:database=' . $CONF->{'db_database'} .
#                           ';host=' . $CONF->{'db_host'} .
#                           ';port=' . $CONF->{'db_port'},
#                           $DB->{'username'}, $DB->{'password'},
#                           {'RaiseError' => 1});
#    $dbh->{FetchHashKeyName} = 'NAME_lc';
#    return $dbh;
#}
#
## Connect to the ldap READ server.
#sub ldap_connect {
#
#    my $ldap;
#    my $ldap_read_host = $CONF->{'ldap_read_host'};
#
#    if ( ($ldap = Net::LDAPapi->new($ldap_read_host)) == $LDAP_CONNECT_FAILED) {
#        croak "ERROR Connection to '$ldap_read_host' failed.";
#    }
#    my $status;
#    if (!$CONF->{'tgt'}) {
#        $status = $ldap->bind_s();
#    } else {
#        $ldap->sasl_parms(-mech=>'GSSAPI');
#        $status = $ldap->bind_s(-type=>LDAP_AUTH_SASL);
#    }
#    if ($status != LDAP_SUCCESS) {
#        if ($ldap) {
#            $ldap->unbind ;
#        }
#        croak "ERROR Bind error connecting to $ldap_read_host";
#    }
#    return $ldap;
#}
#
## Close the connection to the ldap server
#sub ldap_disconnect {
#    my ($ldap) = @_;
#    if ($ldap) {
#        $ldap->unbind;
#    }
#    return;
#}

# Pull the list of the user workgroups from the directory server

sub my_workgroups {
    my %workgroups;

    my $base   = $CONF->{'ldap'}->{'wg_base'};
    my $filter = 'uid='.$ENV{'REMOTE_USER'};
    my $scope  = LDAP_SCOPE_SUBTREE;
    my @attrs  = ('suPrivilegeGroup');

    # Connect to the LDAP.
    $VEMAIL->ldap_read_connect();
    my $ldap = $VEMAIL->ldap_read();

    my $stat = $ldap->search_s($base,$scope,$filter,\@attrs,0);
    if ($stat == LDAP_NO_SUCH_OBJECT) {
        return;
    }

    if ($stat != LDAP_SUCCESS) {
        my $m1 = 'WARNING:'
            . "LDAP search error Base:$base Filter:$filter Scope:$scope";
        my $m2 = 'FATAL:';
        if ($ldap->errstring) {
            $m2 .= $ldap->errstring;
        }
        if ($ldap->extramsg) {
            $m2 .= ' - '.$ldap->extramsg;
        }
        if (!$m2) {
            $m2 = "FATAL: Unknown LDAP error status = $stat";
        }
        croak "$m1\n$m2\n";
    }

    my $entry_ref = $ldap->get_all_entries;

    my %entry = %{ $entry_ref };
    foreach my $dn (keys %entry) {
        foreach my $attr (keys %{$entry{$dn}}) {
            if (lc($attr) eq 'suprivilegegroup') {
                foreach my $pg (@{$entry{$dn}{$attr}}) {
                    $workgroups{$pg} = 1;
                }
            }
        }
    }

    $VEMAIL->ldap_unbind();

    return %workgroups;
}

# Get the list of containers from the database where the container management
# workgroup matches one of user workgroups.  If the user belongs in the admin
# workgroups, then allow the user access to all the containers.
sub my_containers {
    my ($dbh) = @_;
    my %containers;

    # Get a list of suPrivilegeGroups from the directory
    my %groups = my_workgroups();

    # Check to see if this user is an admin
    my $is_admin;
    foreach my $a (@{ $CONF->{'admins'} }) {
        if ($groups{lc($a)}) {
            $is_admin = 1;
            last;
        }
    }

    my $sth1 = $dbh->prepare('SELECT * FROM containers WHERE status = ?');
    my $sth2 = $dbh->prepare('SELECT domain FROM domains WHERE container = ?');

    $sth1->execute('active');
    while (my $row = $sth1->fetchrow_hashref) {
        if ((!$is_admin) and (!$groups{lc($row->{'workgroup'})})) {
            next;
        }
        my $domains = $dbh->selectcol_arrayref($sth2, undef,
                                               $row->{'container'});
        $row->{'domains'} = join (', ', @{ $domains });
        $containers{$row->{'container'}} = $row;
    }
    return %containers;
}

# Verify the input data for a specific alias.  Make sure the given alias is
# alphanumeric plus certain symbols.  Make sure the given maildrops contain
# valid email addresses.  If an owner is given, then make sure the owner is
# of a valid sunetid format.
sub verify_data {
    my ($row, $data, $errors) = @_;

    my $ret = 1;
    if ($data->{'mail'} !~ /^[\w.\-]+$/xsm) {
        my $m = "Row $row: the alias must be the portion of the email "
            . 'address before the &#64; sign.  The alias must be '
            . 'alphanumeric and cannot contain any symbols other than '
            . 'period, underscore, and dash.';
        push @{ $errors }, { message => $m };
        $ret = 0;
    }

    my @maildrops;
    for my $i (0 .. 2) {
        if ($data->{"maildrop$i"}) {
            my @addrs = Email::Address->parse($data->{"maildrop$i"});
            if (@addrs and $addrs[0]->address) {
                push @maildrops, $addrs[0]->address;
            }
        }
    }
    $data->{'maildrop'} = join(q{,}, @maildrops);
    if (!@maildrops) {
        my $m = "Row $row: there are no valid maildrops in this row. "
            . 'Maildrops must be fully qualfied and valid email addresses.';
        push @{ $errors }, { message => $m };
    }

    if ($data->{'owner'}) {
        if ($data->{'owner'} !~ /^[[:lower:]\d]{3,8}$/xsm) {
            my $m = "Row $row: owner must be a SUNetID, 3 to 8 lowercase "
                . 'alphanumeric characters.';
            push @{ $errors }, { message => $m };
            $ret = 0;
        }
    }

    return $ret;
}

# Call the ldap-alias-maint script to synchronize the container data from
# the MySQL database to the ldap server.
sub sync_to_ldap {
    my ($th, $container) = @_;
    my (@errors, @notices);

    my @cmd = () ;
    push (@cmd, '/usr/bin/ldap-alias-maint', 'update', $container) ;

    my ($stdout, $stderr, $rc) = run_command_improved(@cmd);

    if ($stderr) {
        my $m = "Synchronization to LDAP failed: $stderr";
        push(@errors, { message => $m });
        $th->param(ERRORS => \@errors);
        return 0;
    } elsif ($rc) {
        my $m = "Synchronization to LDAP failed: $stdout";
        push(@errors, { message => $m });
        $th->param(ERRORS => \@errors);
        return 0;
    } else {
        my $m = "Container $container successfully synchronized to LDAP.";
        push(@notices, { message => $m });
        $th->param(NOTICES => \@notices);
        return 1;
    }

#    my $res = remctl($CONF->{'remctl_server'}, undef, $CONF->{'remctl_principal'},
#                     'ldapalias', 'update', $container);
#    if ($res->error()) {
#        my $m = 'Synchronization to LDAP failed: ' . $res->error();
#        push @errors, { message => $m };
#        $th->param(ERRORS => \@errors);
#        return 0;
#    } elsif ($res->status()) {
#        my $m = 'Synchronization to LDAP failed: ' . $res->stdout();
#        push @errors, { message => $m };
#        $th->param(ERRORS => \@errors);
#        return 0;
#    } else {
#        my $m = "Container $container successfully synchronized to LDAP.";
#        push @notices, { message => $m };
#        if (@notices) {
#            $th->param(NOTICES => \@notices);
#        }
#        return 1;
#    }
}

##############################################################################
# Action routines
##############################################################################

# Add new aliases to the container after verifying the inputs.  Check
# for duplicate aliases before adding them. Return a list of
# problematic row numbers for further editing.
sub action_container_add {
    my ($dbh, $th, $ch, $params) = @_;
    my (@errors, @notices, @problems);
    my $container = $params->{'container'};

    my $sql1 = 'SELECT * FROM aliases WHERE container = ? AND mail = ?';
    my $sql2 = 'INSERT INTO aliases ('
        . 'container, '
        . 'mail, '
        . 'maildrop, '
        . 'owner, '
        . 'description) '
        . 'values (?, ?, ?, ?, ?)';
    my $sth1 = $dbh->prepare($sql1);
    my $sth2 = $dbh->prepare($sql2);

    my $total = 0;
    foreach my $i (1 .. $NUMBER_OF_ROWS_TO_SHOW) {
        if ($params->{"maildrop0_$i"} || $params->{"maildrop1_$i"} ||
            $params->{"maildrop2_$i"} || $params->{"mail_$i"}) {
            my $data = { 'mail'        => $params->{"mail_$i"},
                         'maildrop0'   => $params->{"maildrop0_$i"},
                         'maildrop1'   => $params->{"maildrop1_$i"},
                         'maildrop2'   => $params->{"maildrop2_$i"},
                         'owner'       => $params->{"owner_$i"},
                         'description' => $params->{"description_$i"} };
            if (!verify_data($i, $data, \@errors)) {
                push(@problems, $i);
                next;
            }
            my $dup = $dbh->selectrow_hashref($sth1, undef, $container,
                                              $data->{'mail'});
            if (defined $dup) {
                my $m = "Row $i: the alias $data->{mail} "
                    . 'already exists in this container.';
                push @errors, { message => $m };
                push(@problems, $i);
                next;
            }
            my $row = $sth2->execute($container,
                                     $data->{'mail'},
                                     $data->{'maildrop'},
                                     $data->{'owner'},
                                     $data->{'description'});
            if ($row == 1) {
                $total++;
            }
        }
    }

    if ($total > 0) {
        push @notices, { message => "Total of $total aliases added." };
    }

    if (@errors) {
        $th->param(ERRORS => \@errors);
    }
    if (@notices) {
        $th->param(NOTICES => \@notices);
    }
    return @problems;
}

# Modify selected aliases in the container after verifying the inputs.  If an
# alias changes the mail attribute, make sure the new mail attribute does not
# already exist in the container.  Return a list of problematic rows for
# further editing.
sub action_container_edit {
    my ($dbh, $th, $ch, $params) = @_;
    my (@errors, @notices, @problems);

    my $container = $params->{'container'};

    my $sql1 = 'SELECT * FROM aliases WHERE container = ? AND mail = ?';
    my $sql2 = 'UPDATE aliases SET '
        . 'mail = ?, '
        . 'maildrop = ?, '
        . 'owner = ?, '
        . 'description = ? '
        . 'WHERE container = ? AND mail = ?';
    my $sth1 = $dbh->prepare($sql1);
    my $sth2 = $dbh->prepare($sql2);

    my $total = 0;
    foreach my $p (keys %{ $params }) {
        if ($p =~ /^mail_(?<rownum>\d+)$/xsm) {
            # Using a named capture group.
            my $i = $LAST_PAREN_MATCH{rownum} . q{};
            my $data = { 'mail'        => $params->{"mail_$i"},
                         'maildrop0'   => $params->{"maildrop0_$i"},
                         'maildrop1'   => $params->{"maildrop1_$i"},
                         'maildrop2'   => $params->{"maildrop2_$i"},
                         'owner'       => $params->{"owner_$i"},
                         'description' => $params->{"description_$i"} };

            if (!verify_data($i, $data, \@errors)) {
                push @problems, $i;
                next;
            }
            if ($params->{"mail_$i"} ne $params->{"old_mail_$i"}) {
                my $dup = $dbh->selectrow_hashref($sth1, undef, $container,
                                                  $data->{'mail'});
                if (defined $dup) {
                    my $m = "Row $i: the alias $data->{mail} "
                        . 'already exists in this container.';
                    push @errors, { message => $m };
                    push @problems, $i;
                    next;
                }
            }
            my $row = $sth2->execute($data->{'mail'},
                                     $data->{'maildrop'},
                                     $data->{'owner'},
                                     $data->{'description'},
                                     $params->{'container'},
                                     $params->{"old_mail_$i"});

            if ($row == 1) {
                $total++;
            }
        }
    }
    if ($total > 0) {
        push @notices, { message => "Total of $total aliases modified." };
    }

    if (@errors) {
        $th->param(ERRORS => \@errors);
    }
    if (@notices) {
        $th->param(NOTICES => \@notices);
    }
    return @problems;
}

# Delete selected aliases from the database.
sub action_container_del {
    my ($dbh, $th, $ch, $container, @aliases) = @_;
    my (@errors, @notices);

    my $total = 0;
    my $sql = 'DELETE FROM aliases WHERE container = ? AND mail = ?';
    my $sth = $dbh->prepare($sql);

    foreach my $alias (@aliases) {
        my $rows = $sth->execute($container, $alias);
        if ($rows >= 1) {
            $total++;
        } else {
            push @errors, { message => "Unable to delete alias $alias." };
        }
    }
    if ($total > 0) {
        push @notices, { message => "Total of $total aliases deleted." };
    }

    if (@errors) {
        $th->param(ERRORS => \@errors);
    }
    if (@notices) {
        $th->param(NOTICES => \@notices);
    }

    return;
}

##############################################################################
# Viewer routines
##############################################################################

# Display an error message if the user does not permission to edit the
# container.
sub view_error_denied {
    my ($th, $ch, $container) = @_;

    $th->param(ERRORS => [ { message => "You do not have permission to edit virtual email container <b>$container</b>." } ] );

    print $ch->header, $th->output;

    return;
}

# Display a page with a list of containers the user has access to.
sub view_containers {
    my ($dbh, $th, $ch, $containers) = @_;

    my $template = HTML::Template->new(filename => 'container.tmpl');

    my @containers = sort { $a->{'container'} cmp $b->{'container'} }
        values %{ $containers };
    if (@containers) {
        $template->param(AT_LEAST_ONE_CONTAINER => '1');
        my $self_url = $ch->url();

        my @container_params = ();
        foreach my $container (@containers) {
            my $cname     = $container->{'container'};
            my $workgroup = $container->{'workgroup'};
            my $contact   = $container->{'contact'};
            my $finance   = $container->{'finance'};
            my $status    = $container->{'status'};
            my $domains   = $container->{'domains'};
            my $url       = "$self_url?container=$cname&list=all";

            push(@container_params,
                {
                    'cname'     => $cname,
                    'workgroup' => $workgroup,
                    'contact'   => $contact,
                    'finance'   => $finance,
                    'status'    => $status,
                    'domains'   => $domains,
                    'url'       => $url,
                }
            ) ;
        }
        $template->param(CONTAINERS => \@container_params);
    } else {
        $template->param(AT_LEAST_ONE_CONTAINER => '0');
    }

    $th->param(CONTENT => $template->output);
    print $ch->header, $th->output;

    return ;
}

# Display the list of aliases in a given container.
sub view_container_list {
    my ($dbh, $th, $ch, $container, $list) = @_;

    my $template = HTML::Template->new(filename => 'aliases.tmpl');
    $template->param(CONTAINER => $container);

    my $sql = 'SELECT * FROM aliases WHERE container = ? ORDER BY mail';
    my $sth = $dbh->prepare($sql);
    $sth->execute($container);
    my $total = 0;
    my @aliases = ();
    while (my $data = $sth->fetchrow_hashref) {
        $total++;
        push(@aliases, $data) ;
        # warn Dumper $data ;
    }
    $template->param(ALIASES => \@aliases);
    $template->param(TOTAL_ALIASES => $total);

    $th->param(CONTENT => $template->output);
    print $ch->header, $th->output;

    return ;

#    my $cnt = h2("Virtual Email Container: $container");
#    $cnt .= start_form(-method => 'GET');
#    $cnt .= hidden('container', $container);
#    $cnt .= submit('add') . '&nbsp;' . submit('delete') . '&nbsp;' .
#            submit('edit') . '&nbsp;' . reset() . submit('Finish') . p();
#    $cnt .= $ch->start_table({-class=>'gold-header'});
#    $cnt .= Tr(th(['Alias', 'Maildrop', 'Owner', 'Description']));
#
#    $sth->execute($container);
#    my $total = 0;
#    while (my $data = $sth->fetchrow_hashref) {
#        $total++;
#        my $checkbox = checkbox(-name  => 'aliases',
#                                -value => $data->{'mail'},
#                                -label => $data->{'mail'});
#        $cnt .= Tr(td([$checkbox,
#                       $data->{'maildrop'},
#                       $data->{'owner'},
#                       $data->{'description'}]));
#    }
#    $cnt .= $ch->end_table();
#    $cnt .= p("$total total aliases");
#    $cnt .= submit('add') . '&nbsp;' . submit('delete') . '&nbsp;' .
#            submit('edit') . '&nbsp;' . reset() . submit('Finish');
#    $cnt .= end_form();
#    $th->param(CONTENT => $cnt);
#    print $ch->header, $th->output;

}

# Display $NUMBER_OF_ROWS_TO_SHOW rows of empty aliases to be added to the
# container unless there are problematic rows from a previous add.
# Display only the problematic rows if there are any problems.
sub view_container_add {
    my ($dbh, $th, $ch, $container, @problems) = @_;

    my $template = HTML::Template->new(filename => 'add-aliases.tmpl');
    $template->param(CONTAINER => $container);

    if (scalar(@problems) == 0) {
        @problems = (1 .. $NUMBER_OF_ROWS_TO_SHOW);
    }

    my @new_aliases = ();
    foreach my $problem (@problems) {
        my $cur_alias = {
            'index' => $problem,
        } ;
        push(@new_aliases, $cur_alias) ;
    }
    $template->param(NEWALIASES => \@new_aliases);

    $th->param(CONTENT => $template->output);
    print $ch->header, $th->output;

    return ;
}

# Display the selected aliases to be edited unless there are
# problematic rows from a previous edit. Only one of
# $aliases, @problems is allowed.
sub view_container_edit {
    my ($dbh, $th, $ch, $container, $aliases_aref, $problems_aref) = @_;
    my (@errors, @confirmed);

    if (defined($aliases_aref) && defined($problems_aref)) {
        croak 'only one of aliases and problems can be defined';
    }

    # This is the array that will be rendered:
    my @edit_entries = () ;

    # If $aliases_aref is defined, get the alias information from the
    # database.
    if (defined($aliases_aref)) {
        # CASE 1: ALIASES
        my @aliases = @{ $aliases_aref };

        my $sql = 'SELECT * FROM aliases WHERE container = ? AND mail = ?';
        my $sth = $dbh->prepare($sql);

        my $index = 0;
        foreach my $alias (@aliases) {
            my $data = $dbh->selectrow_hashref($sth, undef, $container, $alias);
            if (!defined($data)) {
                my $m = "The alias $alias does not exist in this container.";
                push(@errors, { message => $m });
                next;
            }

            ++$index;
            my @maildrops = split(/,/xsm, $data->{'maildrop'});
            my $alias_info = {
                'index'       => $index,
                'mail'        => $data->{'mail'},
                'maildrop0'   => $maildrops[0],
                'maildrop1'   => $maildrops[1],
                'maildrop2'   => $maildrops[2],
                'owner'       => $data->{'owner'},
                'description' => $data->{'description'},
            } ;

            push(@edit_entries, $alias_info) ;
        }
    } else {
        # CASE 2: PROBLEMS
        my @problems = @{ $problems_aref };

        foreach my $problem (@problems) {
            my $alias_info = {
                'index'       => $problem,
                'mail'        => q{},
                'maildrop0'   => q{},
                'maildrop1'   => q{},
                'maildrop2'   => q{},
                'owner'       => q{},
                'description' => q{},
            } ;
            push(@edit_entries, $alias_info) ;
        }
    }

    my $template = HTML::Template->new(filename => 'edit.tmpl');
    $template->param(CONTAINER    => $container);
    $template->param(EDIT_ENTRIES => \@edit_entries);

    $th->param(CONTENT => $template->output);
    print $ch->header, $th->output;

    return;

#    #########################################################################
#
#    my $cnt = h2("Virtual Email Container: $container");
#    $cnt .= h3('Edit existing aliases');
#    $cnt .= start_form(-method => 'POST');
#    $cnt .= hidden('container', $container);
#    $cnt .= submit('confirm edit') . '&nbsp;' . submit('cancel') . p();
#    $cnt .= $ch->start_table({-class=>'gold-header'});
#    $cnt .= Tr(th(['&nbsp;', 'Alias', 'Maildrop', 'Owner', 'Description']));
#
#    my $sql = 'SELECT * FROM aliases WHERE container = ? AND mail = ?';
#    my $sth = $dbh->prepare($sql);
#
#    foreach my $problem (@problems) {
#        $cnt .= Tr(td([
#            $problem . hidden("old_mail_$problem"),
#            textfield(-name => "mail_$problem",
#                      -size => 16,
#                      -maxlength => 64),
#            textfield(-name => "maildrop0_$problem",
#                      -size => 24,
#                      -maxlength => 128) . br() .
#            textfield(-name => "maildrop1_$problem",
#                      -size => 24,
#                      -maxlength => 128) . br() .
#            textfield(-name => "maildrop2_$problem",
#                      -size => 24,
#                      -maxlength => 128),
#            textfield(-name => "owner_$problem",
#                      -size => 8,
#                      -maxlength => 8),
#            textfield(-name => "description_$problem",
#                      -size => 32,
#                      -maxlength => 64)
#        ]));
#    }
#
#    my $i = 0;
#    foreach my $alias (@{ $aliases }) {
#        my $data = $dbh->selectrow_hashref($sth, undef, $container, $alias);
#        if (!defined $data) {
#            my $m = "The alias $alias does not exist in this container.";
#            push @errors, { message => $m };
#            next;
#        }
#        $i++;
#        my @maildrops = split(/,/xsm, $data->{'maildrop'});
#        $cnt .= Tr(td([
#            $i . hidden("old_mail_$i", $data->{'mail'}),
#            textfield(-name => "mail_$i",
#                      -size => 16,
#                      -maxlength => 64,
#                      -default => $data->{'mail'}
#                      ),
#            textfield(-name => "maildrop0_$i",
#                      -size => 24,
#                      -maxlength => 128,
#                      -default => $maildrops[0]) . br() .
#            textfield(-name => "maildrop1_$i",
#                      -size => 24,
#                      -maxlength => 128,
#                      -default => $maildrops[1]) . br() .
#            textfield(-name => "maildrop2_$i",
#                      -size => 24,
#                      -maxlength => 128,
#                      -default => $maildrops[2]),
#            textfield(-name => "owner_$i",
#                      -size => 8,
#                      -maxlength => 8,
#                      -default => $data->{'owner'}),
#            textfield(-name => "description_$i",
#                      -size => 32,
#                      -maxlength => 64,
#                      -default => $data->{'description'})
#        ]));
#    }
#    $cnt .= $ch->end_table();
#    $cnt .= submit('confirm edit') . '&nbsp;' . submit('cancel');
#    $cnt .= end_form();
#
#    if (@errors) {
#        $th->param(ERRORS => \@errors);
#    }
#    $th->param(CONTENT => $cnt);
#    print $ch->header, $th->output;
#
#    return;
}

# Display the selected aliases to be deleted.
sub view_container_del {
    my ($dbh, $th, $ch, $container, @aliases) = @_;
    my (@errors, @confirmed);

    my $template = HTML::Template->new(filename => 'delete-aliases.tmpl');
    $template->param(CONTAINER => $container);

    my @delete_aliases  = () ;
    my @delete_aliases2 = () ;
    my @aliases_params  = () ;

    my $sql = 'SELECT * FROM aliases WHERE container = ? AND mail = ?';
    my $sth = $dbh->prepare($sql);
    my $total = 0;
    foreach my $alias (@aliases) {
        my $data = $dbh->selectrow_hashref($sth, undef, $container, $alias);
        if (!(defined $data)) {
            my $m = "The alias $alias does not exist in this container.";
            push(@errors, { message => $m });
            next;
        }

        my $cur_alias = {
            'mail'        => $data->{'mail'},
            'maildrop'    => $data->{'maildrop'},
            'owner'       => $data->{'owner'},
            'description' => $data->{'description'},
        } ;

        my $mail = $data->{'mail'} ;
        my $cur_alias2 = {
            'mail' => $mail,
        } ;

        push(@delete_aliases,  $cur_alias);
        push(@delete_aliases2, $cur_alias2);
        push(@aliases_params, "aliases=$mail") ;

        $total++;
    }

    $template->param(DELETE_ALIASES  => \@delete_aliases);
    $template->param(DELETE_ALIASES2 => \@delete_aliases2);
    $template->param(TOTAL_ALIASES => $total);

    my $aliases_params = join('&amp;', @aliases_params) ;
    $template->param(ALIASES => $aliases_params);

    if (@errors) {
        $th->param(ERRORS => \@errors);
    }

    $th->param(CONTENT => $template->output);
    print $ch->header, $th->output;

    return;
}

##############################################################################
# Main
##############################################################################

# Get output and errors in the right order.
local $OUTPUT_AUTOFLUSH = 1;

# Read configuration data
$VEMAIL->read_config();
$CONF = $VEMAIL->CONF();

# Connect to the database.
$VEMAIL->database_connect();
my $dbh = $VEMAIL->DB();


# Set ticket for ldapsearches and for remctl
$VEMAIL->kinit();
#refresh_krb5_cache($CONF->{'keytab'}, $CONF->{'tgt'});

# Initialize the template and cgi instance.
my $th = HTML::Template->new(filename => $CONF->{'template'});
my $ch = CGI->new;

# Initialize some common variables.
my %containers = my_containers($dbh);

my $params = $ch->Vars;
my @aliases = $ch->multi_param('aliases');
if ($params->{'container'}) {
    if (exists($containers{$params->{'container'}})) {
        if ($params->{'list'} || $params->{'cancel'}) {
            view_container_list($dbh, $th, $ch, $params->{'container'},
                                $params->{'list'});
        } elsif ($params->{'add'}) {
            view_container_add($dbh, $th, $ch, $params->{'container'});
        } elsif ($params->{'edit'}) {
            view_container_edit($dbh, $th, $ch, $params->{'container'},
                                \@aliases, undef);
        } elsif ($params->{'delete'}) {
            view_container_del($dbh, $th, $ch, $params->{'container'},
                               @aliases);
        } elsif ($params->{'confirm addition'}) {
            my @problems = action_container_add($dbh, $th, $ch, $params);
            if (@problems) {
                view_container_add($dbh, $th, $ch, $params->{'container'},
                                   \@problems);
            } else {
                view_container_list($dbh, $th, $ch, $params->{'container'});
            }
        } elsif ($params->{'confirm edit'}) {
            my @problems = action_container_edit($dbh, $th, $ch, $params);
            if (@problems) {
                view_container_edit($dbh, $th, $ch, $params->{'container'},
                                    undef, \@problems);
            } else {
                view_container_list($dbh, $th, $ch, $params->{'container'});
            }
        } elsif ($params->{'confirm deletion'}) {
            action_container_del($dbh, $th, $ch, $params->{'container'},
                                 @aliases);
            view_container_list($dbh, $th, $ch, $params->{'container'});
        } elsif ($params->{'Finish'}) {
            if (sync_to_ldap($th, $params->{'container'})) {
                view_containers($dbh, $th, $ch, \%containers);
            } else {
                view_container_list($dbh, $th, $ch, $params->{'container'});
            }
        }
    } else {
        view_error_denied($th, $ch, $params->{'container'});
    }
} else {
    view_containers($dbh, $th, $ch, \%containers);
}

# Disconnect from the database and LDAP.
#$dbh->disconnect();
$VEMAIL->database_disconnect();

exit;
